from pathlib import Path
from typing import Dict, Optional

from sqlalchemy.orm import Query
from threedi_schema import ThreediDatabase, models

RASTERS_MAP = [
    ("dem_file", models.ModelSettings),
    ("frict_coef_file", models.ModelSettings),
    ("interception_file", models.Interception),
    # Interflow
    ("porosity_file", models.Interflow),
    ("hydraulic_conductivity_file", models.Interflow),
    # Simple infiltration
    ("infiltration_rate_file", models.SimpleInfiltration),
    ("max_infiltration_capacity_file", models.SimpleInfiltration),
    # Groundwater
    ("initial_infiltration_rate_file", models.GroundWater),
    ("equilibrium_infiltration_rate_file", models.GroundWater),
    ("infiltration_decay_period_file", models.GroundWater),
    ("phreatic_storage_capacity_file", models.GroundWater),
    ("groundwater_impervious_layer_level_file", models.GroundWater),
    ("groundwater_hydro_connectivity_file", models.GroundWater),
    # Initials
    ("initial_waterlevel_file", models.InitialConditions),
    ("initial_groundwater_level_file", models.InitialConditions),
    # Vegetation
    ("vegetation_height_file", models.VegetationDrag2D),
    ("vegetation_drag_coefficient_file", models.VegetationDrag2D),
    ("vegetation_stem_count_file", models.VegetationDrag2D),
    ("vegetation_stem_diameter_file", models.VegetationDrag2D),
]


__all__ = ["ModelDB"]


MIN_SQLITE_VERSION = 300


class ModelDB:
    """Interface to sqlite of a model."""

    def __init__(
        self,
        sqlite_path: Path,
        global_settings_id: Optional[int] = None,
        upgrade: bool = False,
    ):
        if not sqlite_path.exists():
            raise ValueError(f"Sqlite path {sqlite_path} does not exist.")

        self.sqlite_path = sqlite_path
        self.database = ThreediDatabase(self.sqlite_path.as_posix())

        version = self.get_version()
        if version < MIN_SQLITE_VERSION:
            if upgrade:
                self.upgrade()
            else:
                raise ValueError(f"Too old sqlite version {version}.")

        if global_settings_id:
            self.global_settings_id = global_settings_id

        try:
            session = self.database.get_session()
            self.model_settings = (
                session.query(models.ModelSettings).order_by("id").first()
            )
            self.global_settings_name = ""
        finally:
            session.close()

    def get_version(self) -> int:
        # check version
        return self.database.schema.get_version()

    def upgrade(self) -> None:
        self.database.schema.upgrade()

    def get_raster_filepaths(self, base_path: Path) -> Dict:
        session = self.database.get_session()
        raster_filepaths = {}
        for raster, db_table in RASTERS_MAP:
            table_contents = Query(db_table).with_session(session).first()
            raster_path = getattr(table_contents, raster, None)
            if raster_path:
                raster_filepaths[raster] = base_path / Path(
                    raster_path.replace("\\", "/")
                )

        session.close()
        return raster_filepaths

    def initial_waterlevels(self):
        try:
            session = self.database.get_session()
            initial_conditions = (
                Query(models.InitialConditions).with_session(session).first()
            )
            initial_waterlevel = (
                initial_conditions.initial_groundwater_level_aggregation,
                initial_conditions.initial_water_level,
            )
            initial_groundwater_level = (
                initial_conditions.initial_groundwater_level_aggregation,
                initial_conditions.initial_groundwater_level,
            )
        finally:
            session.close()
        return initial_waterlevel, initial_groundwater_level



import logging
from time import sleep

from rich.live import Live
from threedi_api_client.openapi.exceptions import ApiException
from threedi_api_client.openapi.models import Action

from threedi_cmd.console import console
from threedi_cmd.logger import log_settings

from .base import SimulationChildWrapper
from .waitfor import WaitForStatusWrapper, WaitForTimeWrapper

logger = logging.getLogger(__name__)

MAX_RETRIES = 10


class ActionWrapper(SimulationChildWrapper):
    model = Action
    api_path: str = "actions"
    scenario_name = model.__name__.lower()

    def create(self, retry_count: int = 0):
        """Try to create the action, retrying if no more sessions are available"""
        if retry_count > MAX_RETRIES:
            raise Exception("Failed to execute simulation create action after 10 retries")

        try:
            if log_settings.get("use_rich_logging", False):
                with Live(console=console) as live:
                    live.update(f"Creating {self.model.__name__}...")
                    func = getattr(self.api, f"{self.base_path}{self.api_path}_create")

                    data = self.instance.to_dict()
                    data = {key: item for key, item in data.items() if item is not None}

                    res = func(self.simulation.id, data)
                    live.update(
                        f":heavy_check_mark: [bold spring_green4] Created {self.model.__name__}"
                    )
            else:
                func = getattr(self.api, f"{self.base_path}{self.api_path}_create")
                res = func(self.simulation.id, self.instance)
        except ApiException as e:
            if (
                e.status == 429
                and e.body is not None
                and "has no more sessions available" in e.body
            ):
                retry_count += 1
                sleep_time = retry_count**2
                logger.info(
                    f"No more sessions available, retrying... (waiting {sleep_time} seconds)"
                )
                sleep(sleep_time)
                return self.create(retry_count)

            raise e

        console.print(self._data)
        return res

    @property
    def extra_steps(self):
        extra_steps = []
        name = self.instance.name
        if name == "initialize":
            data = {"name": "initialized", "paused": True}
        elif name == "start":
            data = {"name": "initialized", "paused": False}
        elif name == "pause":
            data = {"name": "initialized", "paused": True}
        elif name == "shutdown":
            data = {"name": "finished"}
        elif name == "queue":
            msg = "'queue' step is not yet implemented"
            raise NotImplementedError(msg)
        else:
            msg = f"Unknown name {name}"
            raise ValueError(msg)

        extra = {}
        if "waitfor_timeout" in self._data:
            extra["timeout"] = self._data["waitfor_timeout"]

        wait_for_status = WaitForStatusWrapper(
            data={**data, **extra},
            api_client=self._api_client,
            simulation=self.simulation,
        )
        extra_steps.append(wait_for_status)

        if self.instance.duration is not None:
            extra_steps.append(
                WaitForStatusWrapper(
                    data={"name": "initialized", "paused": True, **extra},
                    api_client=self._api_client,
                    simulation=self.simulation,
                )
            )
            extra_steps.append(
                WaitForTimeWrapper(
                    data={"time": self.instance.duration, **extra},
                    api_client=self._api_client,
                    simulation=self.simulation,
                )
            )
        return extra_steps



import asyncio
import time
from asyncio import Queue
from typing import Dict

from threedi_api_client.openapi.models import (
    ArrivalTimePostProcessing,
    BasicPostProcessing,
    DamagePostProcessing,
    PostProcessingStatus,
)

from threedi_cmd.console import console

from .base import LizardPostprocessingWrapper
from .waitfor import DEFAULT_TIMEOUT, WaitForTimeout


class LizardBasicPostprocessingWrapper(LizardPostprocessingWrapper):
    model = BasicPostProcessing
    api_path: str = "basic"
    scenario_name = "lizardbasicpostprocessing"


class LizardDamagePostprocessingWrapper(LizardPostprocessingWrapper):
    model = DamagePostProcessing
    api_path: str = "damage"
    scenario_name = "lizarddamagepostprocessing"


class LizardArrivalPostprocessingWrapper(LizardPostprocessingWrapper):
    model = ArrivalTimePostProcessing
    api_path: str = "arrival"
    scenario_name = "lizardarrivalpostprocessing"


class WaitForLizardPostprocessingWrapper(LizardPostprocessingWrapper):
    """
    Wait for a Lizard Postprocessing status to be what is specified
    """

    model = PostProcessingStatus
    scenario_name = "waitforlizardpostprocessing"
    extra_fields = ["timeout"]

    def __init__(self, data: Dict, *args, **kwargs):
        if "timeout" in data:
            self.timeout = data.pop("timeout")
        else:
            self.timeout = DEFAULT_TIMEOUT
        super().__init__(data, *args, **kwargs)

    async def execute(self, queue: Queue, steps: list) -> list:
        status = None
        console.rule(
            f'[bold] Waiting for Lizard PostProcessingStatus(status="{self.instance.status}") (timeout: {self.timeout} seconds)',
            style="gold3",
        )
        end_statuses = ["archived", "archiving_failed"]

        start = time.time()
        while True:
            resp = self.api.simulations_results_post_processing_lizard_status_list(
                self.simulation.id
            )
            status = resp.status.lower()
            if status in end_statuses:
                break
            if time.time() - start > self.timeout:
                raise WaitForTimeout(
                    f"Lizard postprocessing result {self.instance.status} not found within timeout"
                )
            await asyncio.sleep(2)

        if status != self.instance.status:
            raise ValueError(f"Expected status {self.instance.status}, got {status}")

        return steps


WRAPPERS = [
    LizardBasicPostprocessingWrapper,
    LizardDamagePostprocessingWrapper,
    LizardArrivalPostprocessingWrapper,
    WaitForLizardPostprocessingWrapper,
]



from __future__ import annotations

import json
import sys
from dataclasses import asdict, dataclass, field
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import List
from urllib.parse import urlparse

import click
import yaml
from rich.prompt import Prompt
from threedi_api_client.aio.threedi_api_client import (
    ThreediApiClient as AsyncThreediApiClient,
)
from threedi_api_client.api import ThreediApi
from threedi_api_client.openapi.exceptions import ApiException
from threedi_api_client.threedi_api_client import Configuration, ThreediApiClient

try:
    from threedi_api_client.auth import get_auth_token, is_token_usable, refresh_api_key
except ImportError:
    from threedi_api_client.threedi_api_client import (
        get_auth_token,
        is_token_usable,
        refresh_api_key,
    )

from threedi_cmd.console import console
from threedi_cmd.errors import ExitCodes

SCENARIO_DIR = Path(__file__).parent.parent / "scenarios"


class EndpointOption(Enum):
    localhost = "http://localhost:8000"
    staging = "https://api.staging.3di.live"
    production = "https://api.3di.live"


@lru_cache()
def get_settings(endpoint):
    return Settings(endpoint=endpoint)


@dataclass
class WebSocketSettings:
    api_base_url: str
    token: str
    host: str = field(init=False)
    proto: str = field(init=False)
    api_version: str = field(init=False)

    def __post_init__(self):
        parsed_url = urlparse(self.api_base_url)
        self.host = parsed_url.netloc
        self.proto = "wss" if parsed_url.scheme == "https" else "ws"
        self.api_version = parsed_url.path.lstrip("/")
        self.token = (
            self.token if self.token.startswith("Bearer") else f"Bearer {self.token}"
        )


@dataclass
class CachedConfig:
    username: str = ""
    access: str = ""
    refresh: str = ""
    organisation_uuid: str = ""
    result_folder: str = ""
    scenario_folder: str = ""

    @classmethod
    def load_from_file(cls, file: Path) -> CachedConfig:
        """
        Loads saved settings from 3di_config.yaml. See save method for available attributes.
        """
        try:
            with open(file, "r") as f:
                cached_data = yaml.load(f, Loader=yaml.FullLoader)
                return cls(**cached_data)
        except OSError:
            # settings file does not yet exist
            return cls()

    def save_to_file(self, file: Path) -> None:
        with open(file, "w") as f:
            yaml.dump(
                asdict(self),
                default_flow_style=False,
                stream=f,
            )


@dataclass
class ScenariosMeta:
    scenario_dir: Path = SCENARIO_DIR

    @property
    def scenarios(self) -> List:
        scenario_list = []
        scenario_files = self.scenario_dir.glob("*.yaml")
        for f in scenario_files:
            with f.open("r") as y:
                content = yaml.load_all(y, Loader=yaml.FullLoader)
                try:
                    meta = next(content)["meta"]
                    meta["file"] = f
                    scenario_list.append(meta)
                except KeyError:
                    raise AttributeError(
                        "The scenario definition is missing the 'meta' section"
                    )
        return scenario_list


class Settings:
    """Settings that are saved (authentication) values between calls to the cli.

    The default settings are saved in the file '3di_config.yaml' which lives in
    the configuration folder of your operating system.
    """

    APP_NAME = "3di-scenario"
    CONFIG_FILE_NAME = "3di_config.yaml"

    def __init__(self, endpoint: str = "", websocket_api_version="v3-beta"):
        self._endpoint = EndpointOption[endpoint]
        self.cached_config = CachedConfig.load_from_file(self.config_file)

        if not is_token_usable(self.access):
            self.credentials_prompt(self.configuration)

        self.websocket_settings = WebSocketSettings(
            api_base_url=self.endpoint + "/" + websocket_api_version, token=self.access
        )

    @property
    def endpoint(self) -> str:
        return self._endpoint.value

    @property
    def username(self) -> str:
        # print(f"prop username {self._username}")
        return self.cached_config.username

    @username.setter
    def username(self, username: str):
        self.cached_config.username = username

    @property
    def access(self) -> str:
        return self.cached_config.access

    @access.setter
    def access(self, access: str):
        self.cached_config.access = access

    @property
    def refresh(self) -> str:
        return self.cached_config.refresh

    @refresh.setter
    def refresh(self, refresh: str):
        self.cached_config.refresh = refresh

    @property
    def organisation_uuid(self) -> str:
        return self.cached_config.organisation_uuid

    @organisation_uuid.setter
    def organisation_uuid(self, organisation_uuid: str):
        self.cached_config.organisation_uuid = organisation_uuid

    @property
    def result_folder(self) -> Path:
        return Path(self.cached_config.result_folder).resolve()

    @result_folder.setter
    def result_folder(self, result_folder: Path):
        if isinstance(result_folder, Path):
            result_folder = result_folder.resolve().as_posix()
        self.cached_config.result_folder = result_folder

    @property
    def scenario_folder(self) -> Path:
        return Path(self.cached_config.scenario_folder).resolve()

    @scenario_folder.setter
    def scenario_folder(self, scenario_folder: Path):
        if isinstance(scenario_folder, Path):
            scenario_folder = scenario_folder.resolve().as_posix()
        self.cached_config.scenario_folder = scenario_folder

    @property
    def app_dir(self) -> Path:
        _app_dir = Path(click.get_app_dir(app_name=self.APP_NAME))
        _app_dir.mkdir(parents=True, exist_ok=True)
        return _app_dir

    @property
    def config_file(self) -> Path:
        return self.app_dir / f"{self._endpoint.name}_{self.CONFIG_FILE_NAME}"

    @property
    def configuration(self) -> Configuration:
        configuration = Configuration(
            host=self.endpoint,
            username=f"{self.username}",
            api_key={
                "Authorization": f"{self.access}",
                "refresh": f"{self.refresh}",
            },
            api_key_prefix={"Authorization": "Bearer"},
        )
        configuration.refresh_api_key_hook = refresh_api_key
        if not self.username:
            self.credentials_prompt(configuration)
        return configuration

    @property
    def api_client(self) -> ThreediApiClient:
        client = ThreediApi(
            config={
                "THREEDI_API_HOST": self.endpoint,
                "THREEDI_API_USERNAME": self.username,
                "THREEDI_API_ACCESS_TOKEN": self.access,
                "THREEDI_API_REFRESH_TOKEN": self.refresh,
            },
            version="v3-beta",
        )
        return client

    @property
    def async_api_client(self) -> AsyncThreediApiClient:
        client = AsyncThreediApiClient(self.configuration)
        return client

    @property
    def scenarios(self) -> List:
        scenarios_meta = ScenariosMeta(self.scenario_folder)
        return scenarios_meta.scenarios

    def save(self) -> None:
        try:
            self.cached_config.save_to_file(self.config_file)
        except IOError as err:
            console.print(
                f"Saving cache to {self.config_file} failed with: {err}", style="error"
            )
            return

    @staticmethod
    def save_settings() -> None:
        ctx = click.get_current_context()
        ctx.obj.save()

    def credentials_prompt(self, config: Configuration) -> None:
        console.rule(
            f":key: Authentication required for {config.host}",
            characters="*",
            style="gold3",
        )
        kwargs = {}
        try:
            if self.username:
                kwargs.update({"default": self.username})
        except AttributeError:
            pass
        username = Prompt.ask("Username", **kwargs)
        password = Prompt.ask("Password", password=True)

        try:
            token = get_auth_token(username, password, config.host)
        except ValueError as err:
            console.print(f":collision: {err}", style="error")
            sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)
        except ApiException as err:
            body = json.loads(err.body)["detail"]
            console.print(
                f"Could not log you in. The server responded with: {body}",
                style="error",
            )
            sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)

        # Update the settings
        self.username = username
        self.password = password
        self.access = token["access"]
        self.refresh = token["refresh"]
        config.api_key = {"Authorization": self.access, "refresh": self.refresh}



import asyncio
import json
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from typing import Dict, List, NoReturn, Optional

from dateutil import parser
from rich.padding import Padding
from rich.progress import (
    BarColumn,
    Progress,
    Task,
    TextColumn,
    TimeRemainingColumn,
)

from threedi_cmd.commands.settings import (
    get_settings,
)
from threedi_cmd.console import console
from threedi_cmd.websockets.clients import WebsocketClient

MAX_LEN_NAME = 30


class StatusColors:
    starting = "gold3"
    initialized = ""
    finished = "green"
    postprocessing = "royal_blue1"
    crashed = "red"
    ended = "dodger_blue1"


class MessageTypes(Enum):
    initial = "active-simulations"
    progress = "progress"
    status = "status"
    new = "active-simulation"


progress = Progress(
    TextColumn(
        "[bold blue]ID {task.fields[sim_id]} | {task.fields[status]}",
        justify="left",
    ),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    TextColumn("[bold cyan]{task.fields[name]}", justify="left"),
    "•",
    TextColumn("[bold]{task.fields[user]}", justify="left"),
    "•",
    TextColumn("[bold dark_goldenrod]{task.fields[organisation]}", justify="left"),
    "•",
    TimeRemainingColumn(),
    console=console,
    auto_refresh=False,
    # transient=True
)


@dataclass
class ProgressInfo:
    # defaults are used for the header,
    # instance overrides these values
    sim_id: str = "ID"
    progress: str = "PROGRESS"
    name: str = "NAME"
    user: str = "USER"
    organisation: str = "ORGANISATION"
    time_reamining: str = "TIME REMAINING"

    @classmethod
    def header(cls):
        return f"Currently running simulations [{cls.sim_id} • {cls.progress} • {cls.name} • {cls.user} • {cls.organisation} • {cls.time_reamining}]"  # noqa


@dataclass
class ProgressTask(ProgressInfo):
    """
    every progress bar is managed through a rich.progress.Task.
    This class holds all task related info, plus the actual
    rich.progress.Task instance called 'task'
    """

    description: str = "running"
    start: bool = field(init=False)
    total: int = 100
    status: str = ""
    task: Task = None
    is_live: bool = False

    _kwarg_fields = {
        "description",
        "sim_id",
        "name",
        "user",
        "organisation",
        "start",
        "total",
        "status",
    }

    def __post_init__(self):
        self.start = True if self.progress else False

    @property
    def kwargs(self) -> Dict:
        d = {field_name: getattr(self, field_name) for field_name in self._kwarg_fields}
        return {k: v for k, v in d.items() if not k.startswith("_")}


class ActiveSimulations:
    STATUS_COLORS = {
        "starting": "gold3",
        "initialized": "",
        "finished": "green",
        "postprocessing": "royal_blue1",
        "crashed": "red",
        "ended": "dodger_blue1",
    }

    def __init__(self, endpoint: str):
        self.endpoint = endpoint  # local | stag | prod
        # local cache {simulation_id: ProgressTask}
        self.tasks = {}

    @cached_property
    def websocket_client(self):
        settings = get_settings(self.endpoint)
        return WebsocketClient(settings=settings.websocket_settings)

    async def add_progress_bar(
        self, progress_task: ProgressTask, progress: Progress
    ) -> None:
        """creates a new progress task and adds a ProgressTask instance to the internal registry"""
        task_id = progress.add_task(**progress_task.kwargs)
        if progress_task.progress:
            progress.update(task_id, advance=int(progress_task.progress))
        task = [t for t in progress.tasks if t.id == task_id][0]
        progress_task.task = task
        self.tasks[int(progress_task.sim_id)] = progress_task

    async def advance_progress(self, data: Dict, progress: Progress) -> None:
        """advances the progress bar for the given tasks by the current simulation progress"""
        simulation_id = data["data"]["simulation_id"]
        sim_progress = data["data"]["progress"]
        progress_task = self.tasks.get(int(simulation_id))
        if not progress_task.task.started:
            progress.start_task(progress_task.task.id)
        if int(sim_progress) <= progress_task.task.percentage:
            return
        advance_by = int(sim_progress) - int(progress_task.task.percentage)
        progress.update(progress_task.task.id, advance=advance_by)
        # auto_refresh is False so do this manually
        progress.refresh()

    @staticmethod
    def is_live_finished(status, progress_task: ProgressTask) -> bool:
        return all(
            (
                status == "crashed",
                progress_task.is_live,
                progress_task.status in ("ended", "postprocessing"),
            )
        )

    async def update_status(self, data: Dict, progress: Progress):
        status = data["data"]["status"]
        simulation_id = data["data"]["simulation_id"]
        progress_task = self.tasks.get(int(simulation_id))
        if self.is_live_finished(status, progress_task):
            status = "finished"
        status_txt = status
        color = self.STATUS_COLORS.get(status)
        if color:
            status_txt = f"[bold {color}]{status_txt}[/bold {color}]"
        update_kwargs = {"status": status_txt}
        advance_by = 100 if status in ("finished", "ended", "postprocessing") else None
        if advance_by and not progress_task.task.started:
            progress.start_task(progress_task.task.id)
            update_kwargs.update({"advance": advance_by})
        progress.update(progress_task.task.id, **update_kwargs)
        # auto_refresh is False so do this manually
        progress.refresh()
        progress_task.status = status

    async def run_monitor(self) -> NoReturn:
        asyncio.ensure_future(self.websocket_client.listen("active-simulations/"))
        q = self.websocket_client.get_queue()
        try:
            await asyncio.wait_for(self.websocket_client.is_connected(), timeout=10)
        except asyncio.TimeoutError:
            console.print("Could not establish WS connection", style="error")
            return
        with progress:
            progress.console.print(Padding("", (1, 0)))
            progress.console.rule(ProgressInfo.header())
            progress.console.print(Padding("", (1, 0)))
            while True:
                data = await q.get()
                message_type = await self.get_msg_type(data)
                if not message_type:
                    continue
                if message_type == MessageTypes.initial:
                    progress_tasks = await self.get_initial_progress_tasks(data)
                    for progress_task in progress_tasks:
                        await self.add_progress_bar(progress_task, progress)
                elif message_type == MessageTypes.progress:
                    await self.advance_progress(data, progress)
                elif message_type == MessageTypes.status:
                    await self.update_status(data, progress)
                elif message_type == MessageTypes.new:
                    progress_task = await self.get_new_progress_task(data)
                    await self.add_progress_bar(progress_task, progress)

    @staticmethod
    async def get_msg_type(data) -> Optional[MessageTypes]:
        try:
            return MessageTypes(data["type"])
        except (KeyError, ValueError):
            return

    @staticmethod
    async def get_new_progress_task(data):
        """
        message from "active-simulations" URI
        type declaration "active-simulation"
        """
        data = data["data"]
        simulation_id = list(data.keys())[0]
        data_json = list(data.values())[0]
        return await ActiveSimulations._get_progress_task(simulation_id, data_json)

    @staticmethod
    async def _get_progress_task(simulation_id, simulation_details) -> ProgressTask:
        details = json.loads(simulation_details)
        if len(details["name"]) > MAX_LEN_NAME:
            name = f"{details['name'][:MAX_LEN_NAME]}..."
        else:
            name = details["name"]

        is_live = int(details["duration"]) == 3153600000
        if is_live:
            dt = parser.parse(details["date_created"])
            name = f"{name} | {dt.strftime('%Y-%m-%d %H:%M')} [bold red][LIVE][/bold red]"
        sim_progress = details["progress"]
        return ProgressTask(
            sim_id=simulation_id,
            name=name,
            user=details["user_name"],
            organisation=details["organisation_name"],
            progress=sim_progress,
            status=details["status"],
            is_live=is_live,
        )

    @staticmethod
    async def get_initial_progress_tasks(data: Dict) -> List[ProgressTask]:
        progress_tasks = []
        for simulation_id, json_details in data["data"].items():
            task_info = await ActiveSimulations._get_progress_task(
                simulation_id, json_details
            )
            progress_tasks.append(task_info)
        return progress_tasks



from pathlib import Path
from unittest.mock import patch

import pytest
from threedi_api_client.openapi import Configuration

from threedi_cmd.commands.settings import CachedConfig, EndpointOption, Settings


@pytest.fixture
def configuration():
    conf = Configuration(
        host="test",
        username="tester",
        api_key={
            "Authorization": "",
            "refresh": "",
        },
        api_key_prefix={"Authorization": "Bearer"},
    )
    return conf


@pytest.fixture
def config_file(tmp_path, cache_content) -> Path:
    c = CachedConfig(**cache_content)
    f = tmp_path / "test_config.yaml"
    c.save_to_file(f)
    yield f


@pytest.fixture
def settings_instance(monkeypatch, configuration, config_file):
    def mock_credentials_prompt(*args, **kwargs):
        return None

    monkeypatch.setattr(Settings, "configuration", configuration)
    monkeypatch.setattr(Settings, "credentials_prompt", mock_credentials_prompt)
    monkeypatch.setattr(Settings, "config_file", config_file)
    yield Settings("localhost")


def test_settings(settings_instance):
    assert settings_instance.username == "test"
    assert (
        settings_instance.access
        == "eyJ0eXAiOiJKV1XPOCJhbGciOiJIUzI1NiJ9.eyJ0b2tlbl90eXBlIjoiYWNjZXNzIiwiZXhwIjoxNjEyNTM3OTk1LCJqdGkiOiI1ZjFlYjQ4ZTU2MTY0OTIzYjk5NzNlYWE3MTZmMDJiMyIsInVzZXJfaWQiOiJyb290In0.OlA9G9_ZIyoK8NViEDEoLFQxqwxLjwmzrio7wAgZHrs"
    )
    assert settings_instance.organisation_uuid == "8df668f217494cdfaff89c1a731f31cb"
    assert settings_instance.endpoint == EndpointOption.localhost.value


def test_set_settings_properties(settings_instance):
    assert settings_instance.username == "test"
    settings_instance.username = "the.great.pretender"
    assert settings_instance.username == "the.great.pretender"
    assert settings_instance.cached_config.username == "the.great.pretender"


def test_save_settigns(config_file, settings_instance):
    assert config_file.exists() and config_file.is_file()
    config_file.unlink()
    assert not config_file.exists()
    settings_instance.save()
    assert config_file.exists() and config_file.is_file()


@patch.object(Settings, "save", side_effect=IOError)
def test_save_settigns_error(config_file, settings_instance):
    with pytest.raises(IOError):
        settings_instance.save()



from . import (
    boundary_conditions,
    breach,
    initial_waterlevels,
    lateral,
    leakage,
    postprocessing,
    rain,
    rasteredit,
    revision,
    savedstate,
    settings,
    sources_sinks,
    structure_control,
    substance,
    waitfor,
    wind,
)
from .actions import ActionWrapper
from .simulation import SimulationWrapper

# Define a list of WRAPPERS
# A wrapper is an object that 'wraps' a certain
# resource of the openapi by automatically mapping
# all of it's openapi client methods.

# For every step in a scenario a wrapper is generated

WRAPPERS = (
    [
        ActionWrapper,
        SimulationWrapper,
    ]
    + rain.WRAPPERS
    + savedstate.WRAPPERS
    + waitfor.WRAPPERS
    + lateral.WRAPPERS
    + sources_sinks.WRAPPERS
    + wind.WRAPPERS
    + breach.WRAPPERS
    + rasteredit.WRAPPERS
    + initial_waterlevels.WRAPPERS
    + structure_control.WRAPPERS
    + boundary_conditions.WRAPPERS
    + leakage.WRAPPERS
    + settings.WRAPPERS
    + revision.WRAPPERS
    + postprocessing.WRAPPERS
    + substance.WRAPPERS
)



import json
from asyncio import Queue

from threedi_api_client.openapi.models import Simulation, Substance, UploadEventFile

from .base import EventWrapper
from .lateral import FileLateralWrapper


class SubstanceWrapper(EventWrapper):
    model = Substance
    api_path: str = "substance"
    scenario_name = model.__name__.lower()

    async def execute(self, queue: Queue, steps: list) -> list:
        """Create the substance and insert the id into the steps"""
        self.api_substance = self.save()
        return self.insert_substance_id_for_steps(steps)

    def save(self) -> Substance:
        """Create the substance"""
        assert isinstance(self.simulation, Simulation)
        return self._api_client.simulations_substances_create(
            self.simulation.id, self.instance
        )

    def insert_substance_id_for_steps(self, steps: list) -> list:
        """
        Substitute the substance id back into the steps.
        Loop over steps and find substance events with the same id as the instance
        """
        new_steps = []
        for sim_step in steps:
            if hasattr(sim_step.instance, "substances") and isinstance(
                sim_step.instance.substances, list
            ):
                for i, substance in enumerate(sim_step.instance.substances):
                    if substance["substance"] == self.instance.id:
                        sim_step.instance.substances[i]["substance"] = (
                            self.api_substance.id
                        )

            if isinstance(sim_step.instance, UploadEventFile):
                self.insert_substance_id_in_upload_event_file(sim_step)

            new_steps.append(sim_step)

        return new_steps

    def insert_substance_id_in_upload_event_file(self, upload_event: FileLateralWrapper):
        """
        Analyse the upload event file and find the substance id to replace.
        Use the copy of the original bulk event file which is initiated by the wrapper.
        """
        assert upload_event.filepath is not None, "wrapper should be initialized first"

        full_path = self._base_path / upload_event.filepath
        with open(full_path, "r") as file:
            data = json.load(file)

        for i, event in enumerate(data):
            if "substances" in event.keys():
                for j, substance in enumerate(event["substances"]):
                    if substance["substance"] == self.instance.id:
                        data[i]["substances"][j]["substance"] = self.api_substance.id

        # Write updated data back to the file
        with open(full_path, "w") as file:
            json.dump(data, file)


WRAPPERS = [SubstanceWrapper]



from dataclasses import dataclass, field
from urllib.parse import urlparse


@dataclass
class WebSocketSettings:
    api_base_url: str
    token: str
    host: str = field(init=False)
    proto: str = field(init=False)
    api_version: str = field(init=False)

    def __post_init__(self):
        parsed_url = urlparse(self.api_base_url)
        self.host = parsed_url.netloc
        self.proto = "wss" if parsed_url.scheme == "https" else "ws"
        self.api_version = parsed_url.path.lstrip("/")



import logging
from functools import lru_cache

from rich.logging import RichHandler

log_settings = {
    "use_rich_logging": True,
}


@lru_cache()
def get_logger(level: str):
    handlers = []
    if log_settings.get("use_rich_logging", False):
        handlers.append(RichHandler(rich_tracebacks=True))

    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=handlers,
    )

    if not log_settings.get("use_rich_logging", False):
        logger = logging.getLogger("rich")
    else:
        logger = logging.getLogger("threedi_cmd")
    return logger


class Dummy(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return self

    def __getattr__(self, *args, **kwargs):
        return self

    def __setattr__(self, *args, **kwargs):
        pass

    def __delattr__(self, *args, **kwargs):
        pass

    def __getitem__(self, *args, **kwargs):
        return self

    def __setitem__(self, *args, **kwargs):
        pass

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        pass



import tempfile
from pathlib import Path

import yaml

from threedi_cmd.schematisation_uploader.yaml_converter import YamlConverter


def test_yaml_converter():
    yaml_converter = YamlConverter()
    content = """
        meta:
            name: "constant rain"
            description: "adds a constant rain event with a duration of 5 minutes at the beginning of the simulation"

        scenario:
            simulation:
                threedimodel: "{{ threedimodel_id }}"
                organisation: "{{ organisation_uuid }}"
                name: "{{ simulation_name }}"
                start_datetime: "{{ datetime_now }}"
                duration: 3600
            steps:
                - constantrain:
                    offset: 0
                    duration: 3600
                    value: 0.00002
                    units: 'm/s'
                - waitforlizardpostprocessing:
                    status: archived
                    timeout: 300
                - substance:
                    name: 'substance1'
                    id: 1
                - waitforstatus:
                    name: 'finished'
                    timeout: 1800
        """
    with tempfile.TemporaryDirectory() as temp_dir:
        yaml_template = Path(temp_dir) / "template.yaml"
        with yaml_template.open("w") as f:
            f.write(content)
        yaml_converter.convert_yaml(yaml_template)

        with open(
            yaml_template.parent / f"{yaml_template.stem}_converted.yaml", "r"
        ) as f:
            converted_content = f.read()
            converted = yaml.load(converted_content, Loader=yaml.FullLoader)

    assert [list(k.keys())[0] for k in converted["scenario"]["steps"]] == [
        "substance",
        "constantrain",
        "action",
        "simulationstatus",
        "waitforlizardpostprocessing",
    ]



import logging
from pathlib import Path
from typing import Any, Callable, List, Type

import threedi_api_client.openapi.models as models
import yaml

from threedi_cmd.models import WRAPPERS

logger = logging.getLogger(__name__)


class Organisation:
    fields = "{{ organisation_uuid }}"


class ThreediModel:
    fields = {
        "schematisation_id": "{{ schematisation_id }}",
        "revision_id": "{{ revision_id }}",
        "auto_update": True,
    }


class Simulation:
    fields = {
        "threedimodel": ThreediModel,
        "organisation": Organisation,
        "name": "{{ simulation_name }}",
        "start_datetime": "{{ datetime_now }}",
        "duration": "{{ duration }}",
    }


scenario_mapping: dict[str, Type[Organisation | ThreediModel | Simulation]] = {
    "organisation": Organisation,
    "simulation": Simulation,
    "threedimodel": ThreediModel,
}


class YamlEvent:
    """Simulation step as defined in the yaml file"""

    def __init__(self, wrapper_model: Any, converters: List[Callable] = []):
        self.model: models = wrapper_model.model
        self.wrapper_model = wrapper_model
        self.converters = (
            wrapper_model.converters if hasattr(wrapper_model, "converters") else []
        )

    def convert_yaml_to_openapi(self, yaml_config: dict, **kwargs: str) -> dict:
        for converter in self.converters:
            yaml_config = converter(yaml_config, **kwargs)

        return self._convert(yaml_config)

    def _convert(self, yaml_config: dict) -> dict:
        # Create placeholders for all required values and fill values if they are present
        # in the attribute_map
        openapi_spec = {}
        for key in self.model.required_fields:
            if key in yaml_config:
                openapi_spec[key] = yaml_config[key]
            else:
                openapi_spec[key] = "{{ " + key + " }}"

        extra_fields = getattr(self.wrapper_model, "extra_fields", [])
        for key, value in yaml_config.items():
            if key in self.model.attribute_map or key in extra_fields:
                openapi_spec[key] = value

        return openapi_spec


event_mapping: dict[str, YamlEvent] = {
    wrapper_model.scenario_name: YamlEvent(wrapper_model) for wrapper_model in WRAPPERS
}


# Order in which events are automatically sorted
# if the event is not in this list, it will be placed at the DEFAULT position
# The events are sorted based on the place in the list.
EVENT_SORT_ORDER = [
    ["substance"],
    "DEFAULT",  # Used when the event is not specified
    ["action"],
    ["simulationstatus"],
    ["waitforlizardpostprocessing"],
]


class YamlConverter:
    """Converts a yaml for a 3Di simulation to a format that can be used by the API"""

    def __init__(self, environment: str = "production"):
        self.environment: str = environment

    def convert_yaml(self, yaml_file: Path) -> None:
        with open(yaml_file, "r") as f:
            data = yaml.load(f, Loader=yaml.FullLoader)

        meta = {}
        if "meta" in data:
            meta = data["meta"]

        scenario = {}
        events = self._get_events(data["scenario"]["steps"])
        scenario["steps"] = self._order_events(events)
        del data["scenario"]["steps"]

        for key, values in data["scenario"].items():
            if key in scenario_mapping:
                model = scenario_mapping[key]
                params = self._extract_scenario(model, values)
                scenario[key] = params
            else:
                logger.warning(f"Could not map unknown scenario: {key}")

        test_name = yaml_file.stem
        with open(yaml_file.parent / f"{test_name}_converted.yaml", "w") as f:
            yaml.dump({"meta": meta, "scenario": scenario}, f)

    def fill_yaml(
        self, yaml_file: Path, schematisation_id: int, revision_id: int
    ) -> None:
        with open(yaml_file, "r") as f:
            data = yaml.load(f, Loader=yaml.FullLoader)

        data["scenario"]["simulation"]["threedimodel"]["schematisation_id"] = (
            schematisation_id
        )
        data["scenario"]["simulation"]["threedimodel"]["revision_id"] = revision_id

        with open(yaml_file, "w") as f:
            yaml.dump(data, f)

    def _extract_scenario(
        self, model: Type[Organisation | ThreediModel | Simulation], values: dict
    ) -> dict | str:
        # Recursivlely extract fields from scenario models
        if isinstance(model.fields, str):
            return model.fields

        params = {}
        for key, v in model.fields.items():
            if key in scenario_mapping:
                params[key] = self._extract_scenario(scenario_mapping[key], {})
            elif key in values:
                params[key] = values[key]
            else:
                params[key] = v

        return params

    def _get_events(self, steps: List[dict]) -> List[dict]:
        events = []
        for step in steps:
            key = list(step.keys())[0]
            if key in event_mapping:
                event: YamlEvent = event_mapping[key]
                config = event.convert_yaml_to_openapi(
                    step[key], **{"environment": self.environment}
                )
                events += [{key: config}]
            else:
                if key != "scenario_from_template":
                    logger.warning(f"Could not map unknown event: {key}")

        events += [{"action": {"name": "start", "waitfor_timeout": 600}}]
        events += [{"simulationstatus": {"name": "finished", "timeout": 600}}]
        return events

    def _order_events(self, events: List[dict]) -> List[dict]:
        # Substance definition events should be executed first

        def get_sort_index(key: str) -> int:
            for i, events in enumerate(EVENT_SORT_ORDER):
                if events != "DEFAULT" and key in events:
                    return i
            return EVENT_SORT_ORDER.index("DEFAULT")

        return sorted(events, key=lambda event: get_sort_index(list(event.keys())[0]))
